#!/usr/bin/env python

import os, sys, re, datetime, logging, json, signal

from kafka import KafkaConsumer
from json import loads
from argparse import ArgumentParser


def _exit(message = None, rc=0):
  if (message): printMsg(message)
  printMsg("Exiting [rc={}]".format(rc))
  sys.exit(rc)

def timeFormat():
  return '%b.%d.%Y-%H.%M.%S.%f'


def printMsg(message):
  print ("[{}] [{}] {}".format(datetime.datetime.now().strftime(timeFormat()), GNode, message))


class MyKafkaConsumer:
    kafka_service = os.environ.get('KAFKA_SERVER')
    kafka_port = os.environ.get('KAFKA_PORT')

    if not kafka_service: kafka_service = "localhost"
    if not kafka_port: kafka_port = "9092"

    server = kafka_service + ':' + kafka_port

    def __init__(self, topic, mregexp, aor, showReadWhenMatchingEnabled):

        # Auto offset reset:
        printMsg("Auto offset reset: '{}'".format(aor))

        # Topic:
        printMsg("Working topic: '{}'".format(topic))

        # Show fulfilled:
        self.showReadWhenMatchingEnabled = showReadWhenMatchingEnabled

        # Json content:
        self.json = None
        if GJfile:
          with open(GJfile) as json_file:
            self.json = json.load(json_file)
          printMsg("Established break condition for expected json ('{}'): {}".format(GJfile, self.json))
        else:
          printMsg("No expected json ('any' selected)")

        # Interpreted as regexp indicator:
        self.mregexp = mregexp
        if GJfile:
          if (self.mregexp): printMsg("Json file will be matched as regular expression")
          else: printMsg("Json file will be matched as exact comparison")


        self.consumer = KafkaConsumer(topic,
                        bootstrap_servers=[self.server],
                        auto_offset_reset=aor,
                        enable_auto_commit=True,
                        value_deserializer=lambda x:
                        json.loads(x.decode('utf-8')))

    def match(self, json):
       if self.mregexp:
         for key, value in  sorted(json.items()):
           match = re.match(self.json[key], value)
           if not match: return False
         return True

       else:
         return (sorted(self.json.items()) == sorted(json.items()))

    def receive_data(self):
        printMsg('Reading data from kafka ...')

        for message in self.consumer:
            # Show read
            if (self.showReadWhenMatchingEnabled if self.json else True):
              printMsg('Data read: {}'.format(message))

            # Break condition (in case that a expected json is blocking consumer):
            if (self.json):
              #if (sorted(self.json.items()) == sorted(message.value.items())):
              if (self.match(message.value)):
                printMsg('Message retrieved: condition fulfilled !')
                return message
              else:
                global GSequence
                GSequence += 1
                #saved='{}.read.{}.json'.format(os.path.basename(GJfile), GSequence)
                saved='{}.read.{}.json'.format(GJfile, GSequence)
                with open(saved, 'w') as outfile:
                  json.dump(message.value, outfile)

                printMsg('Message retrieved: condition not fulfilled (saved {}) !'.format(saved))


    def close (self):
      self.consumer.close()


def handler_function(signum, frame):
  rc=0
  if (GJfile): rc=1
  _exit('Timeout !', rc)


def parse_arguments():

  parser = ArgumentParser(description='Kafka Consumer Simulator')
  parser.add_argument('-n', '--node', help='Node name (spaces are removed)', required=True)
  parser.add_argument('-t', '--topic', help='Kafka topic', required=True)
  parser.add_argument('-o', '--auto-offset-reset', help='Auto offset reset (earliest, latest, etc.). By default: latest', required=False)
  parser.add_argument('-f', '--file', help='Json file to consume (consumer will break execution with rc=0)', required=False)
  parser.add_argument('-m', '--match-regexp', help='Match file as regexp (exact match by default)', required=False, action='store_true')
  parser.add_argument('-w', '--timeout', help='Timeout (in seconds) for completion. Timeout will break execution with rc=1', required=False)
  parser.add_argument('-d', '--debug', help='Debug kafka', required=False, action='store_true')
  parser.add_argument('-r', '--show-read-when-matching-enabled', help='Show read messages. False by default when json file for matching is provided.', required=False, action='store_true')

  arguments = parser.parse_args()

  return arguments


#####################
#      M A I N      #
#####################

if __name__ == "__main__":

  arguments = parse_arguments()

  GNode = arguments.node.replace(" ", "")
  topic = arguments.topic
  auto_offset_reset = arguments.auto_offset_reset
  GJfile = arguments.file
  mregexp = arguments.match_regexp
  timeout = arguments.timeout
  debug = arguments.debug
  show_read_when_matching_enabled = arguments.show_read_when_matching_enabled

  GSequence = 0

  printMsg("Starting")

  if (debug):
    now = datetime.datetime.now().strftime(timeFormat())
    logfile='{}-Consumer-{}.log'.format(GNode, now)
    printMsg("Logging kafka in: '{}'".format(logfile))
    logging.basicConfig(filename=logfile, format='%(asctime)s - %(message)s', datefmt=timeFormat(), level=logging.DEBUG)

  if (timeout):
    printMsg("Established timeout for consumer: '{} seconds'".format(timeout))
    # Signal to terminate after timeout:
    signal.signal(signal.SIGALRM,handler_function)
    signal.alarm(int(timeout))


  if not GJfile:
    if (mregexp): printMsg("To match json file as regexp, you must provide the file. See help (-h) !")
    printMsg("Show messages read (matching is enabled): {}".format(str(show_read_when_matching_enabled)));

  if not auto_offset_reset: auto_offset_reset='latest'

  consumer = None
  try:
    consumer = MyKafkaConsumer(topic, mregexp, auto_offset_reset, show_read_when_matching_enabled)
    consumer.receive_data()
    consumer.close()
    _exit()

  except Exception as e:
    consumer.close()
    _exit(e, 1)
