// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_core_mt_SafeSortedVector_hpp
#define anna_core_mt_SafeSortedVector_hpp

#include <vector>

#include <anna/core/mt/Mutex.hpp>
#include <anna/core/mt/Guard.hpp>
#include <anna/core/util/SortedVector.hpp>

namespace anna {

/**
   Patron para ordenar instancias de objetos en base de una clave. El acceso a la lista de
   datos se realiza desde secciones criticas.

   \param T Clase del patron.
   \param SortBy Clase que ofrece el valor por el que ordenar. Debe implementar un metodo constante con la
    signatura: TKey value (const T*)
   \param TKey Tipo de clave usado para calcular la ordenacion. Debe implementar los operadores '=', '<' y '==' y el
   contructor copia.
*/
template < typename T, typename SortBy, typename TKey = int >
class SafeSortedVector : public SortedVector <T, SortBy, TKey>, public Mutex {
public:
  /**
     Constructor.
  */
  SafeSortedVector() : SortedVector <T, SortBy, TKey> () {;}

  /**
     Constructor copia.
     \param other Instancia de la que copiar.
  */
  explicit SafeSortedVector(const SafeSortedVector& other) : SortedVector <T, SortBy, TKey> (other) {}

  /**
     Devolvera \em true si la instancia recibida como parametro esta contenido en el
     vector o \em en otro caso. Si la instancia recibida es NULL siempre devolvera \em false.
     \param t Instancia a comprobar.
     \return \em true si la instancia recibida como parametro esta contenido en el
     vector o \em en otro caso.
  */
  bool contains(const T* t) const
  throw() {
    if(t == NULL)
      return false;

    Guard guard(this, "SafeSortedVector <T, SortBy, TKey>::contains");
    return  SortedVector <T, SortBy, TKey>::contains(t);
  }

  /**
     Incorpora la instancia recibida como parametro en la lista ordenada de objetos.
     \param t Instancia a guardar en el vector. Si es NULL la operacion no tendra ningun efecto.
     \return \em true si ha registrado la nueva instancia o \em false en otro caso.
  */
  bool add(T* t)
  throw(RuntimeException) {
    if(t == NULL)
      return false;

    Guard guard(this, "SafeSortedVector <T, SortBy, TKey>::add");
    return SortedVector <T, SortBy, TKey>::add(t);
  }

  /**
     Elimina la instancia recibida como parametro de la lista ordenada de objetos.
     \param t Instancia a guardar en el vector. Si es NULL la operacion no tendra ningun efecto.
     \return \em true si ha eliminado la instancia o \em false en otro caso.
  */
  bool erase(T* t)
  throw(RuntimeException) {
    if(t == NULL)
      return false;

    Guard guard(this, "SafeSortedVector <T, SortBy, TKey>::erase");
    return SortedVector <T, SortBy, TKey>::erase(t);
  }

  /**
     Devuelve la instancia asociada a la clave recibida como parametro o NULL si no existe.
     \param key Clave a buscar en el vector.
     \return la instancia asociada a la clave recibida como parametro o NULL si no existe.
  */
  T* find(const TKey key)
  throw() {
    Guard guard(this, "SafeSortedVector <T, SortBy, TKey>::find");
    return SortedVector <T, SortBy, TKey>::find(key);
  }

  /**
     Devuelve la instancia asociada a la clave recibida como parametro o NULL si no existe.
     \param key Clave a buscar en el vector.
     \return la instancia asociada a la clave recibida como parametro o NULL si no existe.
  */
  const T* find(const TKey key) const throw() {
    return const_cast <SafeSortedVector <T, SortBy, TKey>*>(this)->find(key);
  }
};

}

#endif
