// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_core_tracing_TraceWriter_hpp
#define anna_core_tracing_TraceWriter_hpp

#include <vector>

#include <anna/core/mt/NRMutex.hpp>
#include <anna/core/tracing/Logger.hpp>
#include <anna/config/defines.hpp>
#include <anna/core/util/Microsecond.hpp>

namespace anna {

class Configuration;
class TraceWriterObserver;

/**
   Grabador de trazas
   @see Logger
*/
class TraceWriter : public Logger::Writer {
public:
  static const int DefaultMaxKBSize = 4192 * 1024;

  /**
     Constructor.
     \warning Si se usa este constructor las trazas se escribiran en un fichero del directorio temporal
     con el mismo nombre que el pid del proceso.
  */
  TraceWriter();

  /**
     Constructor.
     \param fileName Nombre del fichero de trazas.
     \param maxSize Tama�o maximo del fichero de trazas expresando en bytes.
  */
  TraceWriter(const char* fileName, const int maxSize);

  /**
     Devuelve el tama�o maximo del fichero de trazas.
     \return El tama�o maximo del fichero de trazas.
  */
  int getMaxSize() const throw() { return a_maxSize; }

  /**
     Metodo para cambiar dinamicamente la configuracion del grabador de trazas. Las seccciones
     requeridas para el archivo de configuracion seran:

     \code
     [Trace]
     MaxFileSize = Kbytes maximo de los ficheros de trazas. Por defecto sera 4096 Kbytes
     Level = Nivel de trazas con el que comenzar la configuracion. Por defecto sera Debug para las
     versiones con depuracion y Error para las versiones definitivas.
     Clean = Un 1 indica que los ficheros de trazas usados hasta el momento deben borrarse y un 0
     indica que deben mantenerse. Por defecto sera 1.
     \endcode

     Todos los parametros de la seccion \em Trace son opcionales.

     \param fileName Nombre del fichero de trazas.
     \param configuration Configuracion que contiene las variables y secciones necesarias
     para establecer la configuracion desea del grabador de trazas.

     \warning La configuracion debera estar correctamente cargada. Ver Configuration::load.
  */
  void setup(const char* fileName, const Configuration& configuration) throw();

  /**
     Metodo para cambiar dinamicamente la configuracion del grabador de trazas.

     \param fileName Nombre del fichero de trazas.
     \param maxSize Tama�o maximo del fichero de trazas expresando en bytes. Un valor 0 mantiene
     el valor del tama�o estahlecido hasta el momento.
     \param clean \em true Indica que los ficheros de trazas usados hasta el momento deben
     borrase, \em false indica que los ficheros de trazas deben mantenerse.

     \warning Solo deberia llamarse una vez para cambiar de la configuracion por defecto a la
     configuracion particular de cada proceso.
  */
  void setup(const char* fileName, const int maxSize = 0, const bool clean = true) throw();

  /**
     Saca por pantalla la informacion relevente de este grabador de trazas.
  */
  void printResume() throw();

  /**
     Conecta esta instancia con un TraceWriterObserver, lo que permitira que esta instancia
     conozca cuando se abre o cierra el fichero asociado a la escritura de trazas.

     Este metodo puede invocarse tantas veces como sea necesario con distintas instancias de observador.

     \param observer Instancia del observador que recibira las notificaciones de apertura y
     cierre del fichero usado para sacar las trazas.

     \warning La instancia del TraceWriterObserver debera estar disponible durante toda
     la ejecucion.
  */
  void attach(TraceWriterObserver* observer) throw();

private:
  typedef std::vector <TraceWriterObserver*>::iterator observer_iterator;

  static const int ErrorStream = 2;

  class GuardNoLog {
  public:
    GuardNoLog(NRMutex& mutex) : a_mutex(mutex) { mutex.lock(); }
    ~GuardNoLog() { a_mutex.unlock(); }
  private:
    NRMutex& a_mutex;
  };

  std::string a_outputFile;
  std::string a_outputOldFile;
  int a_maxSize;
  int a_stream;
  Microsecond a_lastTime;
  std::string a_date;
  NRMutex a_mutex;
  std::vector <TraceWriterObserver*> a_observers;
  bool a_observed;

  observer_iterator observer_begin() throw() { return a_observers.begin(); }
  observer_iterator observer_end() throw() { return a_observers.end(); }

  static TraceWriterObserver* observer(observer_iterator& ii) throw() { return *ii; }

  void initialize(const char*) throw() { printResume(); }
  void do_write(int level, const char* text, ...) throw();
  const char* getDate() throw();
  int prepareOutput(const char* date) throw();
};

}

#endif
