// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#ifndef anna_http_wims20_ServerSide_hpp
#define anna_http_wims20_ServerSide_hpp

#include <anna/core/RuntimeException.hpp>
#include <anna/core/util/Tokenizer.hpp>

#include <anna/http/wims20/Abstract.hpp>

namespace anna {

namespace http {

class Request;

namespace wims20 {

/**
   Permite interpretar una URI segn las recomendaciones de WIMS 2.0, lo que facilita
   el desarrollo de aplicaciones integradas en Web 2.0; estas recomendaciones indican
   cmo debe formarse la peticin ServerSide (REpresentational State Transfer) para permitir
   el desarrollo de cualquier servicio.

   Implementa el interfaz WIMS 2.0 desde el punto de vista del proceso que recibe la peticin.

   El formato general de una URI segn la recomendacin de WIMS 2.0 es:

   <p>
http://domain-openapis/path-openapis/serviceID/guid/other_possible_levels?query_parameters
   </p>

   Dnde los campos tienen siguen la siguiente especificacin:
   \li http://domain-openapis: Identifica el recurso del Open API.
   \li path-openapis: Recurso opcional que ajusta la ruta hacia los recursos de ste API.
   \li serviceID: Identificador de recurso.
   \li guid: Identificador del usuario que solicita la peticin.
   \li other_possible_level: Opcionalmente se pueden indicar tantos niveles jerrquicos como fuera
   necesario para el servicio.
   \li query_parameters: Lista de parmetros. Si hay ms de un parmetro se separar con '&'.
*/
class ServerSide : public Abstract {
public:
  /**
   * Contructor indicando el parmetro opcional \em path-openapis. Estos dos parmetros se obtendr como
   * parte de la configuracin de nuestro sistema.
   * \param domain: Identifica el recurso del OpenAPI.
   * \param path: Parmetro opcional que ajusta la ruta hacia los recusos de ste API.
   */
  ServerSide(const std::string& domain, const std::string& path) :
    Abstract("ServerSide", domain, path) {
    for(int ii = 0; ii < SplitCode::Max; ii ++)
      a_tokenizer [ii].activateStrip(true);
  }

  /**
   * Constructor que no usar el parmetro opcional \em path-openapis. Este parmetro se obtendr como
   * parte de la configuracin de nuestro sistema.
   * \param domain: Identifica el recurso del OpenAPI.
   */
  explicit ServerSide(const std::string& domain) :
    Abstract("ServerSide", domain) {
    for(int ii = 0; ii < SplitCode::Max; ii ++)
      a_tokenizer [ii].activateStrip(true);
  }

  /**
   * Obtiene el valor asociado a parmetro recibido como parmetro, que habr sido obtenido
   * de la URI mediante el mtodo #decode.
   * \param name Nombre del parmetro que se desea obtener.
   * \param mode Modo de actuar en caso de que no se encuentre el parmetro.
   * \return El valor asociado al parmetro recibido como parmtro. Puede ser NULL.
   */
  const std::string* getValue(const char* name, const Exception::Mode::_v mode = Exception::Mode::Throw) const throw(RuntimeException);

  /**
   * Obtiene el valor asociado a parmetro recibido como parmetro, que habr sido obtenido
   * de la URI mediante el mtodo #decode.
   * \param name Nombre del parmetro que se desea obtener.
   * \param mode Modo de actuar en caso de que no se encuentre el parmetro.
   * \return El valor asociado al parmetro recibido como parmtro. Puede ser NULL.
   */
  const char* getCStringValue(const char* name, const Exception::Mode::_v mode = Exception::Mode::Throw) const throw(RuntimeException);

  /**
   * Obtiene el valor asociado a parmetro recibido como parmetro, que habr sido obtenido
   * de la URI mediante el mtodo #decode.
   * \param name Nombre del parmetro que se desea obtener.
   * \param mode Modo de actuar en caso de que no se encuentre el parmetro.
   * \return El valor asociado al parmetro recibido como parmtro. Puede ser NULL.
   */
  int getIntegerValue(const char* name, const Exception::Mode::_v mode = Exception::Mode::Throw) const throw(RuntimeException);

  /**
   * Decodifica la URI de la peticin enviada como parmetro y la interpreta segn
   * las recomendaciones de WIMS 2.0
   *
   * \param request Peticin HTTP que va a tratar este servidor.
   */
  void decode(const http::Request& request) throw(RuntimeException);

private:
  struct SplitCode { enum _v { HierarchyAndParameter, HierarchyItem, Parameters, ParameterAndArgument, Max }; };

  void decodeHierarchy(const std::string& hierarchy) throw(RuntimeException);
  void decodeParameters(const std::string& parameters) throw(RuntimeException);

  const Tokenizer& split(const SplitCode::_v splitZone, const std::string&) throw(RuntimeException);

  anna::Tokenizer a_tokenizer [SplitCode::Max];
};

}
}
}

#endif
