// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //

// Standard
#include <stdlib.h> // rand()


#include <anna/diameter.comm/Engine.hpp>
#include <anna/core/tracing/Logger.hpp>
#include <anna/core/tracing/TraceMethod.hpp>
#include <anna/xml/Node.hpp>
#include <anna/comm/Network.hpp>
#include <anna/comm/Host.hpp>
#include <anna/comm/ClientSocket.hpp>
#include <anna/diameter.comm/Transport.hpp>
#include <anna/diameter.comm/Engine.hpp>
#include <anna/diameter.comm/Entity.hpp>
#include <anna/diameter.comm/Server.hpp>
#include <anna/diameter.comm/ClientSession.hpp>
#include <anna/diameter.comm/LocalServer.hpp>
#include <anna/core/functions.hpp>
#include <anna/diameter/internal/sccs.hpp>
#include <anna/diameter.comm/OamModule.hpp>
#include <anna/diameter/codec/functions.hpp>
#include <anna/diameter/helpers/base/functions.hpp>
#include <anna/diameter/helpers/helpers.hpp>
#include <anna/diameter/codec/Message.hpp>
#include <anna/diameter/codec/Avp.hpp>
#include <anna/diameter.comm/Response.hpp>

// STD
#include <map>

using namespace std;
using namespace anna::diameter;


namespace anna {
  namespace diameter {
    namespace stack {
      class Dictionary;
    }
  }
}

comm::Engine::Engine(const char *className, const stack::Dictionary *baseProtocolDictionary) :
  anna::app::Component(className),
  a_autoBind(true),
  a_availableForEntities(false),
  a_availableForLocalServers(false),
  a_cer(true),
  a_dwr(true),
//      a_cea(true),
//      a_dwa(true),
  a_watchdogPeriod(ClientSession::DefaultWatchdogPeriod),
  a_maxConnectionDelay(anna::comm::ClientSocket::DefaultMaxConnectionDelay /* 200 ms*/),
  a_numberOfClientSessionsPerServer(1),
  a_baseProtocolCodecEngine((std::string("baseProtocolCodecEngine_for_") + std::string(className)).c_str(), baseProtocolDictionary)
{
  anna::diameter::sccs::activate();
  a_originRealm = anna::functions::getDomainname();
  a_originHost = anna::functions::getHostname();
  a_ceaPathfile = "";

  // Internal base protocol codec engine:
  a_baseProtocolCodecEngine.setValidationMode(anna::diameter::codec::Engine::ValidationMode::Always); // default was: after decoding
}


void comm::Engine::assertBaseProtocolHealth() throw(anna::RuntimeException) {
  if (!getBaseProtocolCodecEngine()->getDictionary())
    throw anna::RuntimeException("Invalid diameter::comm::Engine object: base protocol dictionary provided on constructor was NULL", ANNA_FILE_LOCATION);
  // it would be interesting to check and identify certain base protocol elements in the dictionary ...
  //  but these things will be checked in runtime and will fail if they should.
}


comm::Server* comm::Engine::allocateServer() throw() { return a_serversRecycler.create(); }
void comm::Engine::releaseServer(Server *server) throw() { a_serversRecycler.release(server); }
comm::ClientSession* comm::Engine::allocateClientSession() throw() { return a_clientSessionsRecycler.create(); }
void comm::Engine::releaseClientSession(ClientSession *clientSession) throw() { a_clientSessionsRecycler.release(clientSession); }


void comm::Engine::setClientCERandDWR(const anna::DataBlock & cer, const anna::DataBlock & dwr) throw(anna::RuntimeException) {
  if(codec::functions::getCommandId(cer) != helpers::base::COMMANDID__Capabilities_Exchange_Request) {
    throw anna::RuntimeException("The message provided as 'CER' is not a Capabilities-Exchange-Request", ANNA_FILE_LOCATION);
  }

  if(codec::functions::getCommandId(dwr) != helpers::base::COMMANDID__Device_Watchdog_Request) {
    throw anna::RuntimeException("The message provided as 'DWR' is not a Device-Watchdog-Request", ANNA_FILE_LOCATION);
  }

  a_cer = cer;
  a_dwr = dwr;
}

void comm::Engine::setClientCERandDWR(const std::string & cer, const std::string & dwr) throw(anna::RuntimeException) {

  // Check for base protocol codec engine health:
  assertBaseProtocolHealth();

  // Build CER
  //   <CER> ::= < Diameter Header: 257, REQ >
  //             { Origin-Host } 264 diameterIdentity
  //             { Origin-Realm } 296 idem
  //          1* { Host-IP-Address } 257, address
  //             { Vendor-Id } 266 Unsigned32
  //             { Product-Name } 269 UTF8String
  //             [Origin-State-Id] 278 Unsigned32
  //           * [ Supported-Vendor-Id ]  265 Unsigned32
  //           * [ Auth-Application-Id ] 258 Unsigned32
  //           * [Acct-Application-Id]  259 Unsigned32
  anna::diameter::codec::Message diameterCER(getBaseProtocolCodecEngine());
  int applicationId = 0 /*anna::diameter::helpers::APPID__3GPP_Rx*/; // Unsigned32
  std::string OH = getOriginHostName();
  std::string OR = getOriginRealmName();
  std::string hostIP = anna::functions::getHostnameIP(); // Address
  int vendorId = anna::diameter::helpers::VENDORID__tgpp; // Unsigned32
  std::string productName = "ANNA Diameter Client"; // UTF8String
  bool encodeDefault = false;

  if (cer != "") {
    try {
      diameterCER.loadXML(cer);
    } catch(anna::RuntimeException &ex) {
      //ex.trace();
      encodeDefault = true;
      LOGWARNING(anna::Logger::warning("CER file not found or unable to parse. Encoding harcoded default version ...", ANNA_FILE_LOCATION));
    }
  }
  else {
    encodeDefault = true;
  }

  if(encodeDefault) {
    diameterCER.setId(anna::diameter::helpers::base::COMMANDID__Capabilities_Exchange_Request);
    diameterCER.setApplicationId(applicationId);
    diameterCER.addAvp(anna::diameter::helpers::base::AVPID__Origin_Host)->getDiameterIdentity()->setValue(OH);
    diameterCER.addAvp(anna::diameter::helpers::base::AVPID__Origin_Realm)->getDiameterIdentity()->setValue(OR);
    diameterCER.addAvp(anna::diameter::helpers::base::AVPID__Host_IP_Address)->getAddress()->fromPrintableString(hostIP.c_str()); // supported by Address class, anyway is better to provide "1|<ip address>"
    diameterCER.addAvp(anna::diameter::helpers::base::AVPID__Vendor_Id)->getUnsigned32()->setValue(vendorId);
    diameterCER.addAvp(anna::diameter::helpers::base::AVPID__Product_Name)->getUTF8String()->setValue(productName);
    diameterCER.addAvp(anna::diameter::helpers::base::AVPID__Auth_Application_Id)->getUnsigned32()->setValue(applicationId);
  }

  // Build DWR
  //   <DWR>  ::= < Diameter Header: 280, REQ >
  //              { Origin-Host }
  //              { Origin-Realm }
  anna::diameter::codec::Message diameterDWR(getBaseProtocolCodecEngine());
  encodeDefault = false;

  if (dwr != "") {
    try {
      diameterDWR.loadXML(dwr);
    } catch(anna::RuntimeException &ex) {
      //ex.trace();
      encodeDefault = true;
      LOGWARNING(anna::Logger::warning("DWR file not found or unable to parse. Encoding harcoded default version ...", ANNA_FILE_LOCATION));
    }
  }
  else {
    encodeDefault = true;
  }

  if(encodeDefault) {
    diameterDWR.setId(anna::diameter::helpers::base::COMMANDID__Device_Watchdog_Request);
    diameterDWR.setApplicationId(applicationId);
    diameterDWR.addAvp(anna::diameter::helpers::base::AVPID__Origin_Host)->getDiameterIdentity()->setValue(OH);
    diameterDWR.addAvp(anna::diameter::helpers::base::AVPID__Origin_Realm)->getDiameterIdentity()->setValue(OR);
  }

  // Assignment for internal encoded versions:
  setClientCERandDWR(diameterCER.code(), diameterDWR.code());
}

void comm::Engine::setWatchdogPeriod(const anna::Millisecond & wp) throw(anna::RuntimeException) {
  if(wp < ClientSession::DefaultWatchdogPeriod) {
    throw anna::RuntimeException(anna::functions::asString("Please set watchdog period over %s", ClientSession::DefaultWatchdogPeriod.asString().c_str()), ANNA_FILE_LOCATION);
  }

  a_watchdogPeriod = wp;
}

void comm::Engine::checkEntityCollision(const socket_v &v) throw(anna::RuntimeException) {
  socket_v::const_iterator it;
  socket_v::const_iterator it_min(v.begin());
  socket_v::const_iterator it_max(v.end());

  for(it = it_min; it != it_max; it++) {
    server_iterator ii = server_find(*it);

    if(ii != server_end())
      throw anna::RuntimeException("diameter::comm::Engine::checkEntityCollision: Server is already reserved by a former created entity. Use another", ANNA_FILE_LOCATION);
  }

  // Check repetitions:
  std::map < socket_t, int/*dummy*/ > auxMap;

  for(it = it_min; it != it_max; it++) auxMap[(*it)] = 0;

  if(auxMap.size() != v.size())
    throw anna::RuntimeException("diameter::comm::Engine::checkEntityCollision: Provided addresses list (sockets) must have all items different", ANNA_FILE_LOCATION);
}

comm::Entity* comm::Engine::createEntity(const socket_v & socketList, const std::string &description)
throw(anna::RuntimeException) {
  Entity* result(NULL);
  anna::Guard guard(this, "anna::diameter::comm::Engine::createEntity");

  if(socketList.size() == 0)
    throw anna::RuntimeException("diameter::comm::Engine::createEntity Address/Port server list provided is empty", ANNA_FILE_LOCATION);

  // Proteccion antes de reservar memoria para una entidad (allocateEntity):
  checkEntityCollision(socketList);

  if((result = allocateEntity()) == NULL)
    throw anna::RuntimeException("diameter::comm::Engine::allocateEntity returns NULL (perhaps virtual method was not implemented)", ANNA_FILE_LOCATION);

  // Initialize:
  result->initialize(); // warning: recycler does not initialize its objects and at least, is important to empty servers vector.
  // Assignments (it could be done at allocate):
  result->setEngine(this); // lo podia haber asignado en el allocateEntity (no importa)
  result->setMaxServers(socketList.size());
  result->setDescription(description);
  entity_key key(getEntityKey(socketList));
  result->a_socketListLiteral = key;
  // Create associated servers:
  socket_v::const_iterator it;
  socket_v::const_iterator it_min(socketList.begin());
  socket_v::const_iterator it_max(socketList.end());
  int count = 1;

  for(it = it_min; it != it_max; it++) {
    result->addServer(*it);

    if(count == 1) result->a_primarySocketLiteral = anna::functions::socketLiteralAsString((*it).first, (*it).second);

    if(count == 2) result->a_secondarySocketLiteral = anna::functions::socketLiteralAsString((*it).first, (*it).second);

    count++;
  }

  a_entities.insert(entity_value_type(key, result));
  LOGDEBUG(
    string msg("diameter::comm::Engine::createEntity | ");
    msg += result->asString();
    msg += anna::functions::asText(" | AutoBind: ", a_autoBind);
    anna::Logger::debug(msg, ANNA_FILE_LOCATION);
  );
  return result;
}


comm::LocalServer *comm::Engine::createLocalServer(const std::string & addr, int port, int maxConnections, const anna::Millisecond & allowedInactivityTime, int category, const std::string & description)
throw(anna::RuntimeException) {
  LocalServer* result(NULL);
  anna::Guard guard(this, "anna::diameter::comm::Engine::createLocalServer");
  // Proteccion antes de reservar memoria para un LocalServer
  socket_t key(addr, port);

  if(a_localServers.find(key) != a_localServers.end())
    throw anna::RuntimeException("LocalServer is already reserved by a former created access point. Cannot create again", ANNA_FILE_LOCATION);

  if((result = allocateLocalServer()) == NULL)
    throw anna::RuntimeException("diameter::comm::Engine::allocateLocalServer returns NULL (perhaps virtual method was not implemented)", ANNA_FILE_LOCATION);

  result->setEngine(this); // lo podia haber asignado en el allocateLocalServer (no importa)
  result->setKey(key);
  result->setCategory(category);
  result->setDescription(description);
  result->setAllowedInactivityTime(allowedInactivityTime);
  result->initializeStatisticResources();
// Los saco con metodos virtuales readXXX del motor:
//   if ((a_cea.isEmpty()) || (a_dwa.isEmpty()))
//      throw anna::RuntimeException("Must define valid CEA and DWA messages by mean setCEAandDWA()", ANNA_FILE_LOCATION);
  a_localServers.insert(localServer_value_type(key, result));
  LOGDEBUG(
    string msg("diameter::comm::Engine::createLocalServer | ");
    msg += result->asString();
    anna::Logger::debug(msg, ANNA_FILE_LOCATION);
  );
//   // Listen: (*)
//   /*if (a_autoListen) */result->enable(); // creates server socket
  result->setMaxConnections(maxConnections); // (*) this enables the listen port ... or not
  return result;
}


comm::Entity* comm::Engine::createEntity(const std::string & addr1, int port1, const std::string & addr2, int port2, const std::string &description)
throw(anna::RuntimeException) {
  socket_v dualList;
  dualList.push_back(socket_t(addr1, port1));
  dualList.push_back(socket_t(addr2, port2));
  return (createEntity(dualList, description));
}


comm::Server* comm::Engine::createServer(Entity *entity, const socket_t & socket)
throw(anna::RuntimeException) {
  Server* result(NULL);
  anna::Guard guard(this, "anna::diameter::comm::Engine::createServer");

  if((result = allocateServer()) == NULL)
    throw anna::RuntimeException("diameter::comm::Engine::allocateServer returns NULL", ANNA_FILE_LOCATION);

  // Initialize:
  result->initialize(); // warning: recycler does not initialize its objects and at least, is important to empty client-sessions vector.
  // Assignments (it could be done at allocate):
  result->a_parent = entity;
  result->a_socket = socket;
  result->setMaxClientSessions(a_numberOfClientSessionsPerServer /* engine */);
  result->a_engine = this;
  result->initializeStatisticResources();

  for(int k = 0; k < a_numberOfClientSessionsPerServer; k++)
    result->addClientSession(k);

  a_servers.insert(server_value_type(socket, result));
//      LOGDEBUG( Lo comento, porque ya se tracea en el createEntity
//         string msg("diameter::comm::Engine::resolveServer | ");
//         msg += result->asString();
//         msg += anna::functions::asText(" | AutoBind: ", a_autoBind);
//         anna::Logger::debug(msg, ANNA_FILE_LOCATION);
//      );
  return result;
}


// Lohacemos privado
comm::ClientSession* comm::Engine::createClientSession(Server *server, int socketId)
throw(anna::RuntimeException) {
  ClientSession* result(NULL);
  anna::Guard guard(this, "anna::diameter::comm::Engine::createClientSession");

  if((result = allocateClientSession()) == NULL)
    throw anna::RuntimeException("diameter::comm::Engine::allocateClientSession returns NULL", ANNA_FILE_LOCATION);

  // Initialize:
  result->initialize(); // warning: recycler does not initialize its objects and at least...
  // Assignments (it could be done at allocate):

  if((a_cer.isEmpty()) || (a_dwr.isEmpty()))
    throw anna::RuntimeException("Must define valid CER and DWR messages by mean setClientCERandDWR()", ANNA_FILE_LOCATION);

  result->a_cer.setBody(a_cer);
  result->a_dwr.setBody(a_dwr);
  result->setWatchdogPeriod(a_watchdogPeriod);
  result->a_parent = server;
  result->a_socketId = socketId;
  result->initializeSequences(); // despues de asignar el server y el socketId (sequences are seed-based by mean exclusive hash)
  result->a_engine = this;
  clientSession_key key = ClientSession::getKey(server->getAddress(), server->getPort(), socketId);
  a_clientSessions.insert(clientSession_value_type(key, result));
//      LOGDEBUG( Lo comento, porque ya se tracea en el createEntity
//         string msg("diameter::comm::Engine::createClientSession | ");
//         msg += result->asString();
//         msg += anna::functions::asText(" | AutoBind: ", a_autoBind);
//         anna::Logger::debug(msg, ANNA_FILE_LOCATION);
//      );
  // Creation:
  anna::comm::Network& network = anna::comm::Network::instantiate();
  result->a_server = network.resolveServer(server->getAddress().c_str(), server->getPort(), true /* autoRecovery */,
                     result->a_receiverFactory, &anna::diameter::comm::Transport::getFactory(),
                     anna::comm::Network::Port::Multiple, anna::comm::Network::DoConnect::No /* (*) */);
  // Delay time on tcp connect:
  result->a_server->setMaxConnectionDelay(a_maxConnectionDelay); // (*)

  // Bind:
  if(a_autoBind) result->bind();

  return result;
}


bool comm::Engine::broadcastEntities(const Message* message) throw(anna::RuntimeException) {
  LOGMETHOD(anna::TraceMethod tttm("diameter::comm::Engine", "broadcastEntities", ANNA_FILE_LOCATION));
  bool allok = true;
  bool ok;

  for(entity_iterator it = entity_begin(), maxii = entity_end(); it != maxii; it ++) {
    try {
      ok = entity(it)->broadcast(message);

      if(!ok) allok = false;
    } catch(anna::RuntimeException &ex) {
      ex.trace();
      allok = false;
    }
  }

  return allok;
}

bool comm::Engine::broadcastLocalServers(const Message* message) throw(anna::RuntimeException) {
  LOGMETHOD(anna::TraceMethod tttm("diameter::comm::Engine", "broadcastLocalServers", ANNA_FILE_LOCATION));
  bool allok = true;
  bool ok;

  for(localServer_iterator it = localServer_begin(), maxii = localServer_end(); it != maxii; it ++) {
    try {
      ok = localServer(it)->broadcast(message);

      if(!ok) allok = false;
    } catch(anna::RuntimeException &ex) {
      ex.trace();
      allok = false;
    }
  }

  return allok;
}

bool comm::Engine::bind() throw(anna::RuntimeException) {
  LOGMETHOD(anna::TraceMethod tttm("diameter::comm::Engine", "bind", ANNA_FILE_LOCATION));
  bool result = true; // all OK return

  for(entity_iterator it = entity_begin(), maxii = entity_end(); it != maxii; it ++) {
    try {
      entity(it)->bind();
    } catch(anna::RuntimeException &ex) {
      ex.trace();
      result = false;
    }
  }

  return result;
}

comm::ClientSession* comm::Engine::findClientSession(const std::string & addr, int port, int socketId, anna::Exception::Mode::_v emode)
throw(anna::RuntimeException) {
  return findClientSession(ClientSession::getKey(addr, port, socketId), emode);
}

comm::ClientSession* comm::Engine::findClientSession(const std::string & key, anna::Exception::Mode::_v emode)
throw(anna::RuntimeException) {
  anna::Guard guard(this, "anna::diameter::comm::Engine::findClientSession");
  clientSession_iterator ii = clientSession_find(key);

  if(ii != clientSession_end())
    return clientSession(ii);

  if(emode != anna::Exception::Mode::Ignore) {
    string msg("diameter::comm::Engine::findClientSession | [addr:port|socketId] = ");
    msg += key;
    msg += " | ClientSession not found";
    anna::RuntimeException ex(msg, ANNA_FILE_LOCATION);

    if(emode == anna::Exception::Mode::Throw)
      throw ex;

    ex.trace();
  }

  return NULL;
}


comm::Server* comm::Engine::findServer(const std::string & addr, int port, anna::Exception::Mode::_v emode)
throw(anna::RuntimeException) {
  anna::Guard guard(this, "anna::diameter::comm::Engine::findServer");
  server_iterator ii = server_find(server_key(addr, port));

  if(ii != server_end())
    return server(ii);

  if(emode != anna::Exception::Mode::Ignore) {
    string msg("diameter::comm::Engine::findServer | addr: ");
    msg += addr;
    msg += anna::functions::asText(" | port: ", port);
    msg += " | Server not found";
    anna::RuntimeException ex(msg, ANNA_FILE_LOCATION);

    if(emode == anna::Exception::Mode::Throw)
      throw ex;

    ex.trace();
  }

  return NULL;
}

comm::Entity* comm::Engine::findEntity(const socket_v & socketList, anna::Exception::Mode::_v emode)
throw(anna::RuntimeException) {
  anna::Guard guard(this, "anna::diameter::comm::Engine::findEntity");
  entity_key key(getEntityKey(socketList));
  entity_iterator ii = entity_find(key);

  if(ii != entity_end())
    return entity(ii);

  if(emode != anna::Exception::Mode::Ignore) {
    string msg("diameter::comm::Engine::findEntity | socket list: ");
    msg += key;
    msg += " | Entity not found";
    anna::RuntimeException ex(msg, ANNA_FILE_LOCATION);

    if(emode == anna::Exception::Mode::Throw)
      throw ex;

    ex.trace();
  }

  return NULL;
}

comm::Entity* comm::Engine::findEntity(const std::string & addr1, int port1, const std::string & addr2, int port2, anna::Exception::Mode::_v emode)
throw(anna::RuntimeException) {
  socket_v dualList;
  dualList.push_back(socket_t(addr1, port1));
  dualList.push_back(socket_t(addr2, port2));
  return (findEntity(dualList, emode));
}


//Entity* Engine::findEntity(int category, anna::Exception::Mode::_v emode)
//throw(anna::RuntimeException) {
//
//   Entity *entity;
//
//   for (entity_iterator it = entity_begin(), maxii = entity_end(); it != maxii; it ++) {
//      entity = entity(it);
//      if (entity->getCategory() == category) return entity;
//   }
//
//   return NULL;
//}


comm::LocalServer* comm::Engine::findLocalServer(const std::string & addr, int port, anna::Exception::Mode::_v emode)
throw(anna::RuntimeException) {
  anna::Guard guard(this, "anna::diameter::comm::Engine::findLocalServer");
  socket_t key(addr, port);
  localServer_iterator ii = localServer_find(key);

  if(ii != localServer_end())
    return localServer(ii);

  if(emode != anna::Exception::Mode::Ignore) {
    string msg("diameter::comm::Engine::findLocalServer | addr: ");
    msg += addr;
    msg += anna::functions::asText(" | port: ", port);
    msg += " | LocalServer not found";
    anna::RuntimeException ex(msg, ANNA_FILE_LOCATION);

    if(emode == anna::Exception::Mode::Throw)
      throw ex;

    ex.trace();
  }

  return NULL;
}


comm::ServerSession* comm::Engine::findServerSession(int socketId, anna::Exception::Mode::_v emode) throw(anna::RuntimeException) {
  anna::Guard guard(this, "anna::diameter::comm::Engine::findServerSession");
  ServerSession *result;

  // Search at each local server:
  for(localServer_iterator it = localServer_begin(), maxii = localServer_end(); it != maxii; it ++) {
    result = localServer(it)->findServerSession(socketId, anna::Exception::Mode::Ignore);

    if(result) return result;
  }

  if(emode != anna::Exception::Mode::Ignore) {
    string msg("diameter::comm::Engine::findServerSession | socketId: ");
    msg += anna::functions::asString(socketId);
    msg += " | ServerSession not found";
    anna::RuntimeException ex(msg, ANNA_FILE_LOCATION);

    if(emode == anna::Exception::Mode::Throw)
      throw ex;

    ex.trace();
  }

  return NULL;
}


void comm::Engine::closeClientSession(comm::ClientSession* clientSession, bool destroy)
throw(anna::RuntimeException) {
  if(clientSession == NULL)
    return;

  LOGDEBUG(
    string msg("diameter::comm::Engine::closeClientSession | ");
    msg += clientSession->asString();
    msg += " | Destroy: ";
    msg += (destroy ? "yes" : "no");
    anna::Logger::debug(msg, ANNA_FILE_LOCATION);
  );
  anna::Guard guard(this, "anna::diameter::comm::Engine::closeClientSession");
  clientSession_iterator ii = clientSession_find(clientSession->getKey());

  if(ii == clientSession_end())
    return;

  try {
    clientSession->setState(ClientSession::State::Closing);

    if(destroy) clientSession->setAutoRecovery(false);

    clientSession->unbind(destroy /* destroy needs to perform immediate close */);

    if(!destroy) return;

    releaseClientSession(clientSession);
  } catch(anna::RuntimeException& ex) {
    ex.trace();
  }

  a_clientSessions.erase(ii);
}




void comm::Engine::closeServer(comm::Server* server, bool destroy)
throw(anna::RuntimeException) {
  if(server == NULL)
    return;

  LOGDEBUG(
    string msg("diameter::comm::Engine::closeServer | ");
    msg += server->asString();
    msg += " | Destroy: ";
    msg += (destroy ? "yes" : "no");
    anna::Logger::debug(msg, ANNA_FILE_LOCATION);
  );
  anna::Guard guard(this, "anna::diameter::comm::Engine::closeServer");
  server_iterator ii = server_find(server->a_socket);

  if(ii == server_end())
    return;

  try {
    server->close(destroy);

    if(!destroy) return;

    releaseServer(server);
  } catch(anna::RuntimeException& ex) {
    ex.trace();
  }

  a_servers.erase(ii);
}


void comm::Engine::closeEntity(comm::Entity* entity, bool destroy)
throw(anna::RuntimeException) {
  if(entity == NULL)
    return;

  LOGDEBUG(
    string msg("diameter::comm::Engine::closeEntity | ");
    msg += entity->asString();
    msg += " | Destroy: ";
    msg += (destroy ? "yes" : "no");
    anna::Logger::debug(msg, ANNA_FILE_LOCATION);
  );
  anna::Guard guard(this, "anna::diameter::comm::Engine::closeEntity");
  entity_iterator ii = entity_find(entity->a_socketListLiteral);

  if(ii == entity_end())
    return;

  try {
    entity->close(destroy);

    if(!destroy) return;

    if(!entity->idle()) { entity->setDeprecated(true); return; }

    releaseEntity(entity);
  } catch(anna::RuntimeException& ex) {
    ex.trace();
  }

  a_entities.erase(ii);
}



void comm::Engine::closeLocalServer(comm::LocalServer* localServer, bool destroy)
throw(anna::RuntimeException) {
  if(localServer == NULL)
    return;

  LOGDEBUG(
    string msg("diameter::comm::Engine::closeLocalServer | ");
    msg += localServer->asString();
    msg += " | Destroy: ";
    msg += (destroy ? "yes" : "no");
    anna::Logger::debug(msg, ANNA_FILE_LOCATION);
  );
  anna::Guard guard(this, "anna::diameter::comm::Engine::closeLocalServer");
  localServer_iterator ii = localServer_find(localServer->getKey());

  if(ii == localServer_end())
    return;

  try {
    localServer->close();

    if(!destroy) return;

    releaseLocalServer(localServer);
  } catch(anna::RuntimeException& ex) {
    ex.trace();
  }

  a_localServers.erase(ii);
}



void comm::Engine::closeEntities(bool destroy) throw(anna::RuntimeException) {
  LOGMETHOD(anna::TraceMethod tttm("diameter::comm::Engine", "closeEntities", ANNA_FILE_LOCATION));
  anna::Guard guard(this, "anna::diameter::comm::Engine::closeEntities");

  for(entity_iterator it = entity_begin(), maxii = entity_end(); it != maxii; it ++)
    closeEntity(entity(it), destroy);
}

void comm::Engine::closeLocalServers(bool destroy) throw(anna::RuntimeException) {
  LOGMETHOD(anna::TraceMethod tttm("diameter::comm::Engine", "closeLocalServers", ANNA_FILE_LOCATION));
  anna::Guard guard(this, "anna::diameter::comm::Engine::closeLocalServers");

  for(localServer_iterator it = localServer_begin(), maxii = localServer_end(); it != maxii; it ++)
    closeLocalServer(localServer(it), destroy);
}

void comm::Engine::eraseDeprecatedIdleEntities() throw() {
  LOGMETHOD(anna::TraceMethod tttm("diameter::comm::Engine", "eraseDeprecatedIdleEntities", ANNA_FILE_LOCATION));
  Entity *et;

  for(entity_iterator it = entity_begin(), maxii = entity_end(); it != maxii; it ++) {
    et = entity(it);

    if(et->isDeprecated() && et->idle()) closeEntity(et, true /* destroy */);
  }
}

int comm::Engine::getOTARequestsForEntities() const throw() {
  int result = 0;

  for(const_entity_iterator it = entity_begin(), maxii = entity_end(); it != maxii; it ++)
    result += entity(it)->getOTARequests();

  return result;
}

int comm::Engine::getOTARequestsForLocalServers() const throw() {
  int result = 0;

  for(const_localServer_iterator it = localServer_begin(), maxii = localServer_end(); it != maxii; it ++)
    result += localServer(it)->getOTARequests();

  return result;
}


void comm::Engine::setOriginRealmName(const std::string & originRealmName) throw() {
  a_originRealm = ((originRealmName != "") ? originRealmName : anna::functions::getDomainname());
}


void comm::Engine::setOriginHostName(const std::string & originHostName) throw() {
  a_originHost = ((originHostName != "") ? originHostName : anna::functions::getHostname());
}



void comm::Engine::raiseAutoRecovery(bool autoRecovery) throw(anna::RuntimeException) {
  LOGMETHOD(anna::TraceMethod tttm("diameter::comm::Engine", "raiseAutoRecovery", ANNA_FILE_LOCATION));

  for(entity_iterator it = entity_begin(), maxii = entity_end(); it != maxii; it ++)
    entity(it)->raiseAutoRecovery(autoRecovery);
}

void comm::Engine::do_stop()
throw() {
  LOGMETHOD(anna::TraceMethod tttm("diameter::comm::Engine", "do_stop", ANNA_FILE_LOCATION));
  close(true /* destroy */);
}

std::string comm::Engine::asString(void) const throw() {
  std::string trace;
  trace =  "\n================================";
  trace += "\nDiameter comm Engine information";
  trace += "\n================================";
  trace += "\nAutoBind: ";
  trace += a_autoBind ? "yes" : "no";
  trace += "\nMaxConnectionDelay: ";
  trace += a_maxConnectionDelay.asString();
  trace += "\nAvailable for entities: ";
  trace += a_availableForEntities ? "yes" : "no";
  trace += "\nAvailable for local servers: ";
  trace += a_availableForLocalServers ? "yes" : "no";
  trace += "\nOTA requests: ";
  trace += anna::functions::asString("%d%s", getOTARequests(), idle() ? " (idle)" : "");
  trace += "\nOTA requests for entities: ";
  trace += anna::functions::asString("%d%s", getOTARequestsForEntities(), idleForEntities() ? " (idle)" : "");
  trace += "\nOTA requests for local servers: ";
  trace += anna::functions::asString("%d%s", getOTARequestsForLocalServers(), idleForLocalServers() ? " (idle)" : "");
  // Entities
  trace += "\nNumber of entities: ";
  trace += anna::functions::asString(a_entities.size());

  for(const_entity_iterator it = entity_begin(); it != entity_end(); it++) {
    trace += "\n";
    trace += entity(it)->asString();
  }

  // Server sockets
  trace += "\nNumber of LocalServers: ";
  trace += anna::functions::asString(a_localServers.size());

  for(const_localServer_iterator it = localServer_begin(); it != localServer_end(); it++) {
    trace += "\n";
    trace += localServer(it)->asString();
  }

  return trace;
}


anna::xml::Node* comm::Engine::asXML(anna::xml::Node* parent) const
throw() {
  parent = anna::app::Component::asXML(parent);
  anna::xml::Node* result = parent->createChild("diameter.comm.Engine");
  result->createAttribute("AutoBind", a_autoBind ? "yes" : "no");
  result->createAttribute("MaxConnectionDelay", a_maxConnectionDelay.asString());
  result->createAttribute("AvailableForEntities", a_availableForEntities ? "yes" : "no");
  result->createAttribute("AvailableForLocalServers", a_availableForLocalServers ? "yes" : "no");
  result->createAttribute("OTArequests", anna::functions::asString("%d%s", getOTARequests(), idle() ? " (idle)" : ""));
  result->createAttribute("OTArequestsForEntities", anna::functions::asString("%d%s", getOTARequestsForEntities(), idleForEntities() ? " (idle)" : ""));
  result->createAttribute("OTArequestsForLocalServers", anna::functions::asString("%d%s", getOTARequestsForLocalServers(), idleForLocalServers() ? " (idle)" : ""));
  result->createAttribute("NumberOfEntities", a_entities.size());
  anna::xml::Node* entities = result->createChild("Engine.Entities");

  for(const_entity_iterator it = entity_begin(); it != entity_end(); it++)
    entity(it)->asXML(entities);

  result->createAttribute("NumberOfLocalServers", a_localServers.size());
  anna::xml::Node* localServers = result->createChild("Engine.LocalServers");

  for(const_localServer_iterator it = localServer_begin(); it != localServer_end(); it++)
    localServer(it)->asXML(localServers);

  // DRA Basics
  // Aspect:
  //  <Engine.RemoteRealm Name="afNodeHostRealm.com">
  //     <Engine.RemoteRealmHost Name="afNodeHostname.afNodeHostRealm.com" ServerSession="localhost:3868|ServerSessionId:4"/>
  //  </Engine.RemoteRealm>
  //  <Engine.RemoteRealm Name="ggsnNodeHostRealm.com">
  //     <Engine.RemoteRealmHost Name="ggsnNodeHostname.ggsnNodeHostRealm.com" ServerSession="localhost:3868|ServerSessionId:6"/>
  //  </Engine.RemoteRealm>
  for (dr_dh_server_sessions_it_t drit = a_dr_dh_server_sessions.begin(); drit != a_dr_dh_server_sessions.end(); drit++) {
    anna::xml::Node* remoteRealm = result->createChild("Engine.RemoteRealm");
    remoteRealm->createAttribute("Name", drit->first);
    dh_server_sessions_map_t *dhServerSessions = (dh_server_sessions_map_t *)&(drit->second);
    for (dh_server_sessions_it_t dhit = dhServerSessions->begin(); dhit != dhServerSessions->end(); dhit++) {
      anna::xml::Node* remoteRealmHost = remoteRealm->createChild("Engine.RemoteRealmHost");
      remoteRealmHost->createAttribute("Name", dhit->first);
      server_sessions_vector_t *serverSessions = (server_sessions_vector_t *)&(dhit->second);
      for (server_sessions_it_t ssit = serverSessions->begin(); ssit != serverSessions->end(); ssit++) {
        std::string socket = anna::functions::socketLiteralAsString((*ssit)->getAddress(), (*ssit)->getPort());
        std::string ss_desc = socket + anna::functions::asString("|ServerSessionId:%d", (*ssit)->getSocketId());
        remoteRealmHost->createAttribute("ServerSession", ss_desc);
      }
    }
  }

  return result;
}

comm::Engine::clientSession_iterator comm::Engine::clientSession_find(const clientSession_key &key) throw() {
  return a_clientSessions.find(key);
}

comm::Engine::server_iterator comm::Engine::server_find(const server_key &key) throw() {
  return a_servers.find(key);
}

comm::Engine::entity_iterator comm::Engine::entity_find(const entity_key &key) throw() {
  return a_entities.find(key);
}

comm::Engine::localServer_iterator comm::Engine::localServer_find(const socket_t &key) throw() {
  return a_localServers.find(key);
}

comm::Engine::entity_key comm::Engine::getEntityKey(const std::string & addr1, int port1, const std::string & addr2, int port2) const throw() {
  socket_v dualList;
  dualList.push_back(socket_t(addr1, port1));
  dualList.push_back(socket_t(addr2, port2));
  return (getEntityKey(dualList));
}

comm::Engine::entity_key comm::Engine::getEntityKey(const socket_v &v) const throw() {
  std::string result;
  socket_v::const_iterator it;
  socket_v::const_iterator it_min(v.begin());
  socket_v::const_iterator it_max(v.end());

  for(it = it_min; it != it_max; it++) {
    result += anna::functions::socketLiteralAsString((*it).first, (*it).second);
    result += " ";
  }

  result.erase(result.size() - 1, 1);  // remove last space
  return result;
}


void comm::Engine::availabilityLostForEntities() throw() {
  a_availableForEntities = false;
  LOGDEBUG(
    std::string msg = "diameter::comm::Engine { Origin-Realm: ";
    msg += getOriginRealmName();
    msg += " | Origin-Host: ";
    msg += getOriginHostName();
    msg += " } has lost its availability for entities";
    anna::Logger::debug(msg, ANNA_FILE_LOCATION);
  );
  // OAM
  OamModule &oamModule = OamModule::instantiate();
  oamModule.activateAlarm(OamModule::Alarm::c_LostAvailabilityOverEntitiesForEngineWithClassName__s__, getClassName());
  oamModule.count(OamModule::Counter::LostAvailabilityOverEngineForEntities);
  // Virtual
  availabilityLostForEntities(this);
}


void comm::Engine::availabilityRecoveredForEntities() throw() {
  a_availableForEntities = true;
  LOGDEBUG(
    std::string msg = "diameter::comm::Engine { Origin-Realm: ";
    msg += getOriginRealmName();
    msg += " | Origin-Host: ";
    msg += getOriginHostName();
    msg += " } has recovered its availability for entities";
    anna::Logger::debug(msg, ANNA_FILE_LOCATION);
  );
  // OAM
  OamModule &oamModule = OamModule::instantiate();
  oamModule.cancelAlarm(OamModule::Alarm::c_LostAvailabilityOverEntitiesForEngineWithClassName__s__, getClassName());
  oamModule.count(OamModule::Counter::RecoveredAvailabilityOverEngineForEntities);
  // Virtual
  availabilityRecoveredForEntities(this);
}


void comm::Engine::availabilityLostForLocalServers() throw() {
  a_availableForLocalServers = false;
  LOGDEBUG(
    std::string msg = "diameter::comm::Engine { Origin-Realm: ";
    msg += getOriginRealmName();
    msg += " | Origin-Host: ";
    msg += getOriginHostName();
    msg += " } has lost its availability for local servers";
    anna::Logger::debug(msg, ANNA_FILE_LOCATION);
  );
  // OAM
  OamModule &oamModule = OamModule::instantiate();
  oamModule.activateAlarm(OamModule::Alarm::c_LostAvailabilityOverLocalServersForEngineWithClassName__s__, getClassName());
  oamModule.count(OamModule::Counter::LostAvailabilityOverEngineForLocalServers);
  // Virtual
  availabilityLostForLocalServers(this);
}


void comm::Engine::availabilityRecoveredForLocalServers() throw() {
  a_availableForLocalServers = true;
  LOGDEBUG(
    std::string msg = "diameter::comm::Engine { Origin-Realm: ";
    msg += getOriginRealmName();
    msg += " | Origin-Host: ";
    msg += getOriginHostName();
    msg += " } has recovered its availability for local servers";
    anna::Logger::debug(msg, ANNA_FILE_LOCATION);
  );
  // OAM
  OamModule &oamModule = OamModule::instantiate();
  oamModule.cancelAlarm(OamModule::Alarm::c_LostAvailabilityOverLocalServersForEngineWithClassName__s__, getClassName());
  oamModule.count(OamModule::Counter::RecoveredAvailabilityOverEngineForLocalServers);
  // Virtual
  availabilityRecoveredForLocalServers(this);
}


bool comm::Engine::refreshAvailabilityForEntities() throw() {
  // Here available
  if(a_availableForEntities) {  // check not-bound state for all client-sessions:
    bool isolate = true;

    for(const_entity_iterator it = entity_begin(); it != entity_end(); it++)
      if(entity(it)->isAvailable()) { isolate = false; break; }

    if(isolate) {
      availabilityLostForEntities();
      return true;
    }

    return false;
  }

  // Here not available
  for(const_entity_iterator it = entity_begin(); it != entity_end(); it++)
    if(entity(it)->isAvailable()) {
      availabilityRecoveredForEntities();
      return true;
    }

  return false;
}

bool comm::Engine::refreshAvailabilityForLocalServers() throw() {
  // Here available
  if(a_availableForLocalServers) {  // check not-bound state for all client-sessions:
    bool isolate = true;

    for(const_localServer_iterator it = localServer_begin(); it != localServer_end(); it++)
      if(localServer(it)->isAvailable()) { isolate = false; break; }

    if(isolate) {
      availabilityLostForLocalServers();
      return true;
    }

    return false;
  }

  // Here not available
  for(const_localServer_iterator it = localServer_begin(); it != localServer_end(); it++)
    if(localServer(it)->isAvailable()) {
      availabilityRecoveredForLocalServers();
      return true;
    }

  return false;
}


void comm::Engine::readDPA(anna::DataBlock &dpa, const anna::DataBlock & dpr) throw() {

  // Check for base protocol codec engine health:
  try {
    assertBaseProtocolHealth();
  }
  catch(anna::RuntimeException &ex) {
    ex.trace();
    return;
  }

  // Default DPA implementation:
  //
  //   'Disconnect-Peer-Answer' (282,answer)
  //        {Result-Code}...................................(268,0)
  //        {Origin-Host}...................................(264,0)
  //        {Origin-Realm}..................................(296,0)
  //        [Error-Message].................................(281,0)
  //       *[Failed-AVP]....................................(279,0)
  try {
    anna::diameter::codec::Message diameterDPA(getBaseProtocolCodecEngine());
    anna::diameter::codec::Avp avpRC(getBaseProtocolCodecEngine());
    anna::diameter::codec::Avp avpOH(getBaseProtocolCodecEngine());
    anna::diameter::codec::Avp avpOR(getBaseProtocolCodecEngine());
    // Message header
    diameterDPA.setId(anna::diameter::helpers::base::COMMANDID__Disconnect_Peer_Answer);
    diameterDPA.setVersion(1);
    diameterDPA.setApplicationId(codec::functions::getApplicationId(dpr));
    diameterDPA.setHopByHop(codec::functions::getHopByHop(dpr));
    diameterDPA.setEndToEnd(codec::functions::getEndToEnd(dpr));
    // Result-Code
    avpRC.setId(anna::diameter::helpers::base::AVPID__Result_Code);
    avpRC.setMandatoryBit();
    avpRC.getUnsigned32()->setValue(helpers::base::AVPVALUES__Result_Code::DIAMETER_SUCCESS);
    // Origin-Host
    avpOH.setId(anna::diameter::helpers::base::AVPID__Origin_Host);
    avpOH.setMandatoryBit();
    avpOH.getDiameterIdentity()->fromPrintableString(a_originHost.c_str());
    // Origin-Realm
    avpOR.setId(anna::diameter::helpers::base::AVPID__Origin_Realm);
    avpOR.setMandatoryBit();
    avpOR.getDiameterIdentity()->fromPrintableString(a_originRealm.c_str());
    diameterDPA.addAvp(&avpRC);
    diameterDPA.addAvp(&avpOH);
    diameterDPA.addAvp(&avpOR);
    // Encode
    dpa = diameterDPA.code();
  } catch(anna::RuntimeException &ex) {
    std::string msg = ex.getText();
    msg += " | Use diameter::comm::Engine::setBaseProtocolCodecEngine() to allow internal base protocol messages encoding (unable to answer with DPA)";
    anna::Logger::error(msg, ANNA_FILE_LOCATION);
    //throw anna::RuntimeException(msg, ANNA_FILE_LOCATION);
  }
}


void comm::Engine::readCEA(anna::DataBlock &cea, const anna::DataBlock &cer) throw() {

  // Check for base protocol codec engine health:
  assertBaseProtocolHealth();

  if (a_ceaPathfile != "") {
    anna::diameter::codec::Message diameterCEA(getBaseProtocolCodecEngine());

    try {
      diameterCEA.loadXML(a_ceaPathfile);
      diameterCEA.setHopByHop(anna::diameter::codec::functions::getHopByHop(cer));
      diameterCEA.setEndToEnd(anna::diameter::codec::functions::getEndToEnd(cer));
      cea = diameterCEA.code();

    } catch(anna::RuntimeException &ex) {
      ex.trace();
      LOGWARNING(anna::Logger::warning("CEA file not found or unable to parse. Encoding harcoded default version ...", ANNA_FILE_LOCATION));
      //return anna::diameter::comm::Engine::readCEA(cea, cer);
      // will fail with empty cea
      }

    return;
  }

  // Default CEA implementation:
  //
  //   'Capabilities-Exchange-Answer' (257,answer)
  //        {Result-Code}...................................(268,0)
  //        {Origin-Host}...................................(264,0)
  //        {Origin-Realm}..................................(296,0)
  //      1*{Host-IP-Address}...............................(257,0)
  //        {Vendor-Id}.....................................(266,0)
  //        {Product-Name}..................................(269,0)
  //        [Origin-State-Id]...............................(278,0)
  //        [Error-Message].................................(281,0)
  //       *[Failed-AVP]....................................(279,0)
  //       *[Supported-Vendor-Id]...........................(265,0)
  //       *[Auth-Application-Id]...........................(258,0)
  //       *[Inband-Security-Id]............................(299,0)
  //       *[Acct-Application-Id]...........................(259,0)
  //        [Vendor-Specific-Application-Id]................(260,0)
  //        [Firmware-Revision].............................(267,0)
  //       *[AVP]...........................................(0,0)
  try {
    anna::diameter::codec::Message diameterCEA(getBaseProtocolCodecEngine());
    anna::diameter::codec::Avp avpRC(getBaseProtocolCodecEngine());
    anna::diameter::codec::Avp avpOH(getBaseProtocolCodecEngine());
    anna::diameter::codec::Avp avpOR(getBaseProtocolCodecEngine());
    // Message header
    diameterCEA.setId(anna::diameter::helpers::base::COMMANDID__Capabilities_Exchange_Answer);
    diameterCEA.setVersion(1);
    diameterCEA.setApplicationId(codec::functions::getApplicationId(cer));
    diameterCEA.setHopByHop(codec::functions::getHopByHop(cer));
    diameterCEA.setEndToEnd(codec::functions::getEndToEnd(cer));
    // Result-Code
    avpRC.setId(anna::diameter::helpers::base::AVPID__Result_Code);
    avpRC.setMandatoryBit();
    avpRC.getUnsigned32()->setValue(helpers::base::AVPVALUES__Result_Code::DIAMETER_SUCCESS); // re-implementations could analyze CER to accept or not
    // Origin-Host
    avpOH.setId(anna::diameter::helpers::base::AVPID__Origin_Host);
    avpOH.setMandatoryBit();
    avpOH.getDiameterIdentity()->fromPrintableString(a_originHost.c_str());
    // Origin-Realm
    avpOR.setId(anna::diameter::helpers::base::AVPID__Origin_Realm);
    avpOR.setMandatoryBit();
    avpOR.getDiameterIdentity()->fromPrintableString(a_originRealm.c_str());
    diameterCEA.addAvp(&avpRC);
    diameterCEA.addAvp(&avpOH);
    diameterCEA.addAvp(&avpOR);
    // Host-IP-Address
    std::string hostIP = anna::functions::getHostnameIP(); // Address
    diameterCEA.addAvp(anna::diameter::helpers::base::AVPID__Host_IP_Address)->getAddress()->fromPrintableString(hostIP.c_str());
    // Vendor-Id
    int vendorId = anna::diameter::helpers::VENDORID__tgpp; // Unsigned32
    diameterCEA.addAvp(anna::diameter::helpers::base::AVPID__Vendor_Id)->getUnsigned32()->setValue(vendorId);
    // Product-Name
    std::string productName = "Diameter Server"; // UTF8String
    diameterCEA.addAvp(anna::diameter::helpers::base::AVPID__Product_Name)->getUTF8String()->setValue(productName);
    // Encode
    cea = diameterCEA.code();
  } catch(anna::RuntimeException &ex) {
    std::string msg = ex.getText();
    msg += " | Use diameter::comm::Engine::setBaseProtocolCodecEngine() to allow internal base protocol messages encoding (unable to answer with CEA)";
    anna::Logger::error(msg, ANNA_FILE_LOCATION);
    //throw anna::RuntimeException(msg, ANNA_FILE_LOCATION);
  }
}

void comm::Engine::manageDrDhServerSession(ServerSession *ss, bool register_or_desregister) throw() {

  // Decode CER (TODO: use raw buffer helpers)
  std::string destinationRealm, destinationHost;
  codec::Message codecMsg(getBaseProtocolCodecEngine());
  try {
    codecMsg.decode(ss->a_cer);
    destinationRealm = codecMsg.getAvp(anna::diameter::helpers::base::AVPID__Origin_Realm)->getDiameterIdentity()->getValue();
    destinationHost = codecMsg.getAvp(anna::diameter::helpers::base::AVPID__Origin_Host)->getDiameterIdentity()->getValue();
  }
  catch(anna::RuntimeException &ex) {
    ex.trace();
    return;
  }

  dr_dh_server_sessions_nc_it_t drit = a_dr_dh_server_sessions.find(destinationRealm);
  if (drit != a_dr_dh_server_sessions.end()) { // found
    dh_server_sessions_map_t *dhServerSessions = (dh_server_sessions_map_t *)&(drit->second);
    dh_server_sessions_nc_it_t dhit = dhServerSessions->find(destinationHost);
    if (dhit != dhServerSessions->end()) { // found
      server_sessions_vector_t *serverSessions = (server_sessions_vector_t *)&(dhit->second);
      if (register_or_desregister) { // REGISTER
        serverSessions->push_back(ss);
      }
      else { // DESREGISTER
        // Sequential search the specific server session:
        for (server_sessions_nc_it_t ssit = serverSessions->begin(); ssit != serverSessions->end(); ssit++) {
          if ((*ssit)->getAddress() != ss->getAddress()) continue;
          if ((*ssit)->getPort() != ss->getPort()) continue;
          if ((*ssit)->getSocketId() != ss->getSocketId()) continue;
          serverSessions->erase(ssit);  // if it is the last server session removed in DR-DH path, the XML will show this tree empty
                                        // (it could be a hint for past registerings):
                                        //          <Engine.RemoteRealm Name="afNodeHostRealm.com">
                                        //             <Engine.RemoteRealmHost Name="afNodeHostname.afNodeHostRealm.com"/>
                                        //          </Engine.RemoteRealm>
                                        //          <Engine.RemoteRealm Name="ggsnNodeHostRealm.com">
                                        //             <Engine.RemoteRealmHost Name="ggsnNodeHostname.ggsnNodeHostRealm.com"/>
                                        //          </Engine.RemoteRealm>

          break;
        }
      }
    }
    else {
      if (!register_or_desregister) return; // strange (host not found)
      server_sessions_vector_t ssVector;
      ssVector.push_back(ss);
      (*dhServerSessions)[destinationHost] = ssVector;
    }
  }
  else {
    if (!register_or_desregister) return; // strange (realm not found)
    server_sessions_vector_t ssVector;
    ssVector.push_back(ss);
    dh_server_sessions_map_t dhServerSessions;
    dhServerSessions[destinationHost] = ssVector;
    a_dr_dh_server_sessions[destinationRealm] = dhServerSessions;
  }
}

void comm::Engine::readDWA(anna::DataBlock &dwa, const anna::DataBlock & dwr) throw() {

  // Check for base protocol codec engine health:
  assertBaseProtocolHealth();

  // Default DWA implementation:
  //
  //   'Device-Watchdog-Answer' (280,answer)
  //        {Result-Code}...................................(268,0)
  //        {Origin-Host}...................................(264,0)
  //        {Origin-Realm}..................................(296,0)
  //        [Error-Message].................................(281,0)
  //       *[Failed-AVP]....................................(279,0)
  //        [Origin-State-Id]...............................(278,0)
  try {
    anna::diameter::codec::Message diameterDWA(getBaseProtocolCodecEngine());
    anna::diameter::codec::Avp avpRC(getBaseProtocolCodecEngine());
    anna::diameter::codec::Avp avpOH(getBaseProtocolCodecEngine());
    anna::diameter::codec::Avp avpOR(getBaseProtocolCodecEngine());
    // Message header
    diameterDWA.setId(anna::diameter::helpers::base::COMMANDID__Device_Watchdog_Answer);
    diameterDWA.setVersion(1);
    diameterDWA.setApplicationId(codec::functions::getApplicationId(dwr));
    diameterDWA.setHopByHop(codec::functions::getHopByHop(dwr));
    diameterDWA.setEndToEnd(codec::functions::getEndToEnd(dwr));
    // Result-Code
    avpRC.setId(anna::diameter::helpers::base::AVPID__Result_Code);
    avpRC.setMandatoryBit();
    avpRC.getUnsigned32()->setValue(helpers::base::AVPVALUES__Result_Code::DIAMETER_SUCCESS);
    // Origin-Host
    avpOH.setId(anna::diameter::helpers::base::AVPID__Origin_Host);
    avpOH.setMandatoryBit();
    avpOH.getDiameterIdentity()->fromPrintableString(a_originHost.c_str());
    // Origin-Realm
    avpOR.setId(anna::diameter::helpers::base::AVPID__Origin_Realm);
    avpOR.setMandatoryBit();
    avpOR.getDiameterIdentity()->fromPrintableString(a_originRealm.c_str());
    diameterDWA.addAvp(&avpRC);
    diameterDWA.addAvp(&avpOH);
    diameterDWA.addAvp(&avpOR);
    // Encode
    dwa = diameterDWA.code();
  } catch(anna::RuntimeException &ex) {
    std::string msg = ex.getText();
    msg += " | Use diameter::comm::Engine::setBaseProtocolCodecEngine() to allow internal base protocol messages encoding (unable to answer with DWA)";
    anna::Logger::error(msg, ANNA_FILE_LOCATION);
    //throw anna::RuntimeException(msg, ANNA_FILE_LOCATION);
  }
}

void comm::Engine::resetStatistics() throw() {
  for(server_iterator it = server_begin(), maxii = server_end(); it != maxii; it ++)
    server(it)->resetStatistics();

  for(localServer_iterator it = localServer_begin(), maxii = localServer_end(); it != maxii; it ++)
    localServer(it)->resetStatistics();
}

void comm::Engine::do_initialize() throw(RuntimeException) {
  LOGMETHOD(anna::TraceMethod tttm("diameter::comm::Engine", "do_initialize", ANNA_FILE_LOCATION));
  LOGDEBUG(anna::Logger::debug("Nothing special done on component initialization", ANNA_FILE_LOCATION));
}

void comm::Engine::lazyInitialize() throw(RuntimeException) {
  LOGMETHOD(anna::TraceMethod tttm("diameter::comm::Engine", "lazyInitialize", ANNA_FILE_LOCATION));
  anna::app::Component::initialize(); // this will invoke do_initialize
}

// Not tested yet
const comm::Response* comm::Engine::sendRealmHost(const Message* message, const std::string &destinationRealm, const std::string &destinationHost) throw(anna::RuntimeException) {

  if (destinationRealm == "")
    throw anna::RuntimeException("Unable to resolve the destination: empty provided Destination-Realm name", ANNA_FILE_LOCATION);

  // Get the server sessions which fulfill the restrictions:
  dr_dh_server_sessions_it_t drit = a_dr_dh_server_sessions.find(destinationRealm);
  if (drit == a_dr_dh_server_sessions.end())
    throw anna::RuntimeException(anna::functions::asString("Unable to resolve the destination: Destination-Realm name is not registered (no remote clients have been connected to '%s')", destinationRealm.c_str()), ANNA_FILE_LOCATION);

  dh_server_sessions_map_t *dhServerSessions = (dh_server_sessions_map_t *)&(drit->second);
  // randomize between all server sessions for all hosts:
  dh_server_sessions_nc_it_t dhit;
  int hostsN = dhServerSessions->size();
  if (hostsN == 0) // avoids next division by cero (rand() % 0)
    throw anna::RuntimeException(anna::functions::asString("Unable to resolve the destination: neither Destination-Host currently connected to Destination-Realm '%s'", destinationRealm.c_str()), ANNA_FILE_LOCATION);

  if (destinationHost == "") {
    // in this case, randomize the host:
    dhit = dhServerSessions->begin();
    int randomHostIndx = rand() % hostsN; // number between 0 and the number of hosts - 1
    std::advance (dhit, randomHostIndx);
  }
  else {
    dhit = dhServerSessions->find(destinationHost);
    if (dhit == dhServerSessions->end())
      throw anna::RuntimeException(anna::functions::asString("Unable to resolve the destination: Destination-Host '%s' is not registered for Destination-Realm '%s'", destinationHost.c_str(), destinationRealm.c_str()), ANNA_FILE_LOCATION);
  }

  // Now, randomize the available server sessions:
  server_sessions_vector_t *serverSessions = (server_sessions_vector_t *)&(dhit->second);
  int serverSessionN = serverSessions->size();
  if (serverSessionN == 0) { // avoids next division by cero (rand() % 0)
    std::string aux = "";
    if (destinationHost != "") { aux = "to Destination-Host '"; aux += destinationHost; aux += "'"; }
    std::string msg = anna::functions::asString("Unable to resolve the destination: neither server session currently connected%s within Destination-Realm '%s'", aux.c_str(), destinationRealm.c_str());
    throw anna::RuntimeException(msg, ANNA_FILE_LOCATION);
  }

  server_sessions_nc_it_t ssit = serverSessions->begin();
  int randomServerSessionIndx = rand() % serverSessionN; // number between 0 and the number of server sessions - 1
  std::advance (ssit, randomServerSessionIndx);
  return (*ssit)->send(message);
}
