// ANNA - Anna is Not Nothingness Anymore                                                         //
//                                                                                                //
// (c) Copyright 2005-2015 Eduardo Ramos Testillano & Francisco Ruiz Rayo                         //
//                                                                                                //
// See project site at http://redmine.teslayout.com/projects/anna-suite                           //
// See accompanying file LICENSE or copy at http://www.teslayout.com/projects/public/anna.LICENSE //


#include <anna/core/mt/Guard.hpp>

#include <anna/xml/Compiler.hpp>
#include <anna/xml/Node.hpp>
#include <anna/xml/Attribute.hpp>
#include <anna/xml/Text.hpp>
#include <anna/xml/internal/sccs.hpp>
#include <anna/xml/Document.hpp>
#include <anna/xml/Namespace.hpp>

using namespace anna;
using namespace anna::xml;

Compiler::Compiler() {
  sccs::activate();
}

const char* Compiler::apply(const Node* node, const int flags)
throw(RuntimeException) {
  Guard guard(this, "xml::Compiler::apply");
  a_result.clear();

  if(flags & Mode::Compact)
    apply(node, a_result, flags);
  else
    apply(node, a_result, 0, flags);

  return a_result.c_str();
}

const char* Compiler::apply(const Document& document, const Node* node, const int flags)
throw(RuntimeException) {
  Guard guard(this, "xml::Compiler::apply");
  a_result.clear();
  const char* version = document.getVersion();
  a_result += "<?xml";

  if(version != NULL) {
    a_result += " version='";
    a_result += version;
    a_result += "'";
  }

  const char* encoding = document.getEncoding();

  if(encoding != NULL) {
    a_result += " encoding='";
    a_result += encoding;
    a_result += "'";
  }

  a_result += "?>\n";

  if(flags & Mode::Compact)
    apply(node, a_result, flags);
  else
    apply(node, a_result, 0, flags);

  return a_result.c_str();
}

//------------------------------------------------------------------------------------
// (1) Si tiene nodos hijos => ignora el texto, porque no se como representar
// nodos + texto.
//------------------------------------------------------------------------------------
void Compiler::apply(const Node* node, Result& result, const int level, const int flags)
throw(RuntimeException) {
  const bool hasText(node->getText() != NULL);
  Node::const_child_iterator ii = node->child_begin();
  Node::const_child_iterator maxii = node->child_end();
  const bool hasChildren(ii != maxii);

  if(hasChildren == false && hasText == false)
    open(node, result, level, true, true, flags);
  else if(hasChildren == false && hasText == true) {
    open(node, result, level, false, false, flags);
    result += node->getText()->getValue();
    close(node, result, 0, flags);
  } else if(hasChildren == true) {                    // (1)
    open(node, result, level, false, true, flags);

    for(; ii != maxii; ii ++)
      apply(Node::node(ii), result, level + 1, flags);

    close(node, result, level, flags);
  }
}

//------------------------------------------------------------------------------------
// (1) Si tiene nodos hijos => ignora el texto, porque no se como representar
// nodos + texto.
//------------------------------------------------------------------------------------
void Compiler::apply(const Node* node, Result& result, const int flags)
throw(RuntimeException) {
  static const int level = 0;
  const bool hasText(node->getText() != NULL);
  Node::const_child_iterator ii = node->child_begin();
  Node::const_child_iterator maxii = node->child_end();
  const bool hasChildren(ii != maxii);

  if(hasChildren == false && hasText == false)
    open(node, result, level, true, false, flags);
  else if(hasChildren == false && hasText == true) {
    open(node, result, level, false, false, flags);
    result += node->getText()->getValue();
    result += "</";
    writeFullName(node, result, flags);
    result += ">";
  } else if(hasChildren == true) {                    // (1)
    open(node, result, level, false, false, flags);

    for(; ii != maxii; ii ++)
      apply(Node::node(ii), result, flags);

    result += "</";
    writeFullName(node, result, flags);
    result += ">";
  }
}

/*static*/
void Compiler::open(const Node* node, Result& result, const int level, const bool quickClose, const bool newline, const int flags)
throw(RuntimeException) {
  const Attribute* attribute;
  const Namespace* ns;

  for(int i = 0; i < level; i ++)
    result += ANNA_XML_COMPILER_TAB;

  result += '<';
  writeFullName(node, result, flags);

  // Si es el nodo ROOT => Tenemos que representar la deficin de los namespaces
  if(node->getParent() == NULL && node->namespace_size() != 0) {
    for(xml::Node::const_namespace_iterator ii = node->namespace_begin(), maxii = node->namespace_end(); ii != maxii; ii ++) {
      ns = xml::Node::xnamespace(ii);
      result += " xmlns:";
      result += ns->getName();
      result += "=\"";
      result += ns->getReference();
      result += "\"";
    }
  }

  for(Node::const_attribute_iterator ii = node->attribute_begin(), maxii = node->attribute_end(); ii != maxii; ii ++) {
    attribute = Node::attribute(ii);
    result += ' ';
    writeFullName(attribute, result, flags);
    const std::string& value = attribute->getValue();

    if(value.length() > 0) {
      result += "=\"";
      result += attribute->getValue();
      result += "\"";
    } else if(flags & Mode::ShowNullAttribute)
      result += "=\"\"";
  }

  if(quickClose == true)
    result += '/';

  result += '>';

  if(newline == true)
    result += '\n';
}

/*static*/
void Compiler::writeFullName(const Node* node, Result& result, const int flags)
throw(RuntimeException) {
  const Namespace* ns;

  if((flags & Mode::NoNamespaces) == 0 && (ns = node->getNamespace()) != NULL) {
    result += ns->getName();
    result += ':';
  }

  result += node->getName();
}

/*static*/
void Compiler::writeFullName(const Attribute* attr, Result& result, const int flags)
throw(RuntimeException) {
  const Namespace* ns;

  if((flags & Mode::NoNamespaces) == 0 && (ns = attr->getNamespace()) != NULL) {
    result += ns->getName();
    result += ':';
  }

  result += attr->getName();
}

/*static*/
void Compiler::close(const Node* node, Result& result, const int level, const int flags)
throw(RuntimeException) {
  for(int i = 0; i < level; i ++)
    result += ANNA_XML_COMPILER_TAB;

  result += "</";
  writeFullName(node, result, flags);
  result += ">\n";
}

void Compiler::Result::extend(const int nbytes)
throw(RuntimeException) {
  if((a_size + nbytes) >= a_maxSize) {
    int newSize = ((a_size + nbytes) << 1) - ((a_size + nbytes)  >> 1);
    char* newBuffer = new char [newSize];

    if(newBuffer == NULL)
      throw RuntimeException("Insufficient memory", ANNA_FILE_LOCATION);

    if(a_size > 0)
      anna_memcpy(newBuffer, a_buffer, a_size);

    delete [] a_buffer;
    a_buffer = newBuffer;
    a_maxSize = newSize;
  }
}

